#!/usr/bin/env python

# future improvements
# 1) It'd be nice to have a mode of operation that will close and reopen on every n reads or n writes or n reads and writes, so that we don't have to hold exclusive access to the resource forever
import sys
import time

try:
	import uniq
except:
	have_uniq_module=0
else:
	have_uniq_module=1

class database_cache_item:
	def __init__(self, value):
		self.value = value
		self.use_time = -1

	def mark_used(self):
		self.use_time = time.time()

	def mark_unused(self):
		self.use_time = -1

#	def is_used_recently(self):
#		return self.used_recently

class database_cache:
	def __init__(self, databasefile, databasetype, databasemode, to_string=None, from_string=None, max_elements_in_memory=10000, read_use=True, write_use=True, write_through=True, too_many_percent=95.0, \
			verbose=False):
		""" Module implements both write-through and write-back caching for a python database like anydbm or gdbm """
		self.databasefile = databasefile
		self.database = databasetype.open(self.databasefile, databasemode)
		self.to_string = to_string
		self.from_string = from_string
		self.max_elements_in_memory = max_elements_in_memory
		self.read_use = read_use
		self.write_use = write_use
		self.write_through = write_through
		self.cache = {}
		self.number_elements_in_memory = 0
		self.too_many_percent = too_many_percent
		self.verbose = verbose
#		if self.verbose:
#			sys.stderr.write('read_use is %s, write_use is %s\n' % (self.read_use, self.write_use))

	def flush_if_too_many(self):
		# empty the cache of any items that haven't been used recently, if the cache is more than "too_many_percent" recently used
#		if self.verbose:
#			sys.stderr.write('In flush_if_too_many\n')
		if self.number_elements_in_memory >= self.max_elements_in_memory:
			if self.verbose:
				sys.stderr.write('Flushing some items to disk - starting with %d elements_in_memory - ' % self.number_elements_in_memory)
			# start from the end and work toward the beginning
			keys = self.cache.keys()
			num_keys = self.number_elements_in_memory
			minimum_time = -1
			maximum_time = -1
			if not keys:
				# we have an empty cache - don't continue looking for things to remove from the cache :)  This'll simplify the logic below slightly, too
				return
			# this loop is just to avoid needing low and high sentinels.  Instead, we only use -1 as an undefined sentinel
			for key in keys:
				if minimum_time == -1 and self.cache[key].use_time != -1:
					minimum_time = self.cache[key].use_time
				if maximum_time == -1 and self.cache[key].use_time != -1:
					maximum_time = self.cache[key].use_time
				if minimum_time != -1 and maximum_time != -1:
					break
			if minimum_time == -1 or maximum_time == -1:
				# everything in the cache is equally, 100% non-recent, so toast it all and return :)
				self.cache = {}
				self.number_elements_in_memory = 0
				return
			# yes, we have two unnecessary compares here (due to the sentinel avoidance above), but that's far better than slicing the list
			# anyway, this list is finding the highest use_time and lowest use_time
			for key in keys:
				if self.cache[key].use_time != -1:
					if self.cache[key].use_time < minimum_time:
						minimum_time = self.cache[key].use_time
					if self.cache[key].use_time > maximum_time:
						maximum_time = self.cache[key].use_time
			#sys.stderr.write('minimum_time is %f, maximum_time is %f\n' % (minimum_time, maximum_time))
			# note that the list number in this for loop, 100, should as a last resort remove 100% of the items in the cache - and the final 99-100% will be removed one a time, though it
			# could be done en masse
			# The point of this loop, is to try eliminating the oldest 10%, then 20% ... 90%, followed by 91%, 92% ... 100%
			# Note that these percentages being thrown around are based on time, -not- the number of elements in the cache!  The former is an inexpensive computation, while the latter would
			# be a bit lengthy
			for percentage in range(10,100,10) + range(90,101,1):
			#for percentage in range(1,101,1):
				if self.verbose:
					sys.stderr.write('%d%% ' % percentage)
				cutoff_time = ((maximum_time - minimum_time) * percentage / 100.0) + minimum_time
				for key in self.cache.keys():
					if self.cache[key].use_time < cutoff_time:
						if self.to_string == None:
							self.database[key] = self.cache[key].value
						else:
							self.database[key] = self.to_string(self.cache[key].value)
						del self.cache[key]
						num_keys -= 1
				#sys.stderr.write('%d < %f? ' % (num_keys, self.max_elements_in_memory * self.too_many_percent / 100.0))
				if num_keys < self.max_elements_in_memory * self.too_many_percent / 100.0:
					break
			self.number_elements_in_memory = num_keys
			if self.verbose:
				sys.stderr.write('ending with %d elements_in_memory\n' % self.number_elements_in_memory)
			
	def __getitem__(self, key):
		if self.cache.has_key(key):
			if self.read_use:
				self.cache[key].mark_used()
#			if self.verbose:
#				sys.stderr.write('cachedb %s, 1, number_elements_in_memory is %d, max_elements_in_memory is %d\n' % (self.databasefile, self.number_elements_in_memory, self.max_elements_in_memory))
			return self.cache[key].value
		else:
			if self.database.has_key(key):
				if self.from_string == None:
					self.cache[key] = database_cache_item(self.database[key])
				else:
					self.cache[key] = database_cache_item(self.from_string(self.database[key]))
				self.number_elements_in_memory += 1
				if self.read_use:
					self.cache[key].mark_used()
				self.flush_if_too_many()
#				if self.verbose:
#					sys.stderr.write('cachedb %s, 2, number_elements_in_memory is %d, max_elements_in_memory is %d\n' % (self.databasefile, self.number_elements_in_memory, self.max_elements_in_memory))
				return self.cache[key].value
			else:
				raise KeyError

	def __setitem__(self, key, value):
		if self.cache.has_key(key):
			if self.write_use:
				self.cache[key].mark_used()
			self.cache[key] = database_cache_item(value)
			if self.write_through:
				if self.to_string == None:
					self.database[key] = self.cache[key].value
				else:
					self.database[key] = self.to_string(self.cache[key].value)
		else:
			self.cache[key] = database_cache_item(value)
			if self.write_use:
				self.cache[key].mark_used()
			self.number_elements_in_memory += 1
			if self.to_string == None:
				self.database[key] = self.cache[key].value
			else:
				self.database[key] = self.to_string(self.cache[key].value)
			self.flush_if_too_many()
#		if self.verbose:
#			sys.stderr.write('cachedb %s, 3, number_elements_recently_used is %d, max_elements_in_memory is %d\n' % (self.databasefile, self.number_elements_in_memory, self.max_elements_in_memory))

	def has_key(self, key):
		return self.cache.has_key(key) or self.database.has_key(key)

	def keys(self):
		# possible this behavior choice should be a tunable option
		if have_uniq_module:
			# assuming good hash distribution, O(n), otherwise things could get really ugly!
			result = uniq.uniq(self.cache.keys() + self.database.keys())
		else:
			# O(n*log(n)), unfortunately, and that's actually a pretty decent duration for this task
			k1 = self.cache.keys()
			k2 = self.database.keys()
			k1.sort()
			k2.sort()
			result = []
			k1_ind = 0
			k2_ind = 0
			lk1 = len(k1)
			lk2 = len(k2)
			while k1_ind < lk1 and k2_ind < lk2:
				if k1[k1_ind] < k2[k2_ind]:
					result.append(k1[k1_ind])
					k1_ind += 1
				elif k1[k1_ind] > k2[k2_ind]:
					result.append(k2[k2_ind])
					k2_ind += 1
				else:
					result.append(k1[k1_ind])
					k1_ind += 1
					k2_ind += 2
			for finish_up in range(k1_ind,lk1):
				result.append(k1[finish_up])
			for finish_up in range(k2_ind,lk2):
				result.append(k2[finish_up])
		return result

	def flush(self):
		# If this is a "write through" cache, then we have nothing that needs to be done before termination.  But if this is a "write back" cache, we need to get
		# all that stuff out of the cache and onto disk!
		if not self.write_through:
			for key in self.cache.keys():
				if self.to_string == None:
#					print type(self.cache[key].value)
					self.database[key] = self.cache[key].value
					del self.cache[key]
				else:
					self.database[key] = self.to_string(self.cache[key].value)
					del self.cache[key]

	sync = flush

	def close(self):
		self.flush()
		self.database.close()

if __name__ == '__main__':
	import sys
	import bsddb
	database_type=bsddb
	#import gdbm
	#database_type=gdbm
	import string

	def sequence():
		for i in xrange(100):
			yield i

	cached_db = database_cache('/tmp/foobar.cached', database_type, 'w', str, string.atoi, max_elements_in_memory=5)
	uncached_db = database_type.open('/tmp/foobar.uncached', 'w')
	# test for accuracy, but not performance
	for i in sequence():
		sys.stderr.write('Writing %d\n' % i)
		cached_db[str(i)] = i
		uncached_db[str(i)] = str(i)
	cached_db.flush()
	for i in sequence():
		sys.stderr.write('Comparing %d\n' % i)
		a = cached_db[str(i)]
		b = string.atoi(uncached_db[str(i)])
		if a != b:
			sys.stderr.write('Error: item %d does not match: %s, %s\n' % (i, a, b))
			sys.exit(1)
	sys.stderr.write('All test elements compared correctly :)\n')

